import time
from inspect import currentframe, getouterframes
from collector import SSH
from UtilsRPE.Functions import load_vault
from consistency_checks import check_consistency, saving_state, checkpoint, recover_checkpoint, EmptyQueuesError

import sys

sys.path.insert(0, '/home/Earth/otintopr/RPE-Canal/AutoRPE')
from constructor_of_worlds import set_difference, set_1_contained_in_2, Job


def print_time():
    line = getouterframes(currentframe())[1][2]
    print("Line:%i   Time:%s" % (line, time.time()))


class Queue:
    def __init__(self):
        self.pending = []
        self.running = []
        self.completed = []
        self.suspended = []

    def add(self, element):
        # Proper place?
        element.is_successful = None
        element.submitted = False
        element.is_completed = False
        element._result = None

        if element in self.completed:
            self.completed.remove(element)
            print("Relaunching %s  that was already in COMPLETED" % element.identifier())
        if element in self.running:
            element.cancel_remote()
            self.running.remove(element)
            print("Element %s moved from RUNNING" % element.identifier())
        if element not in self.pending:
            print("Adding element %s that has %i variables" % (element.identifier(), len(element.variable_set)))
            self.pending.insert(0, element)
        else:
            print("\tElement %s already in queue" % (element.identifier()))

    def add_to_end(self, element):
        self.pending.append(element)

    def first(self):
        element = self.pending.pop(0)
        print("\tDelivering %s from queue" % (element.identifier()))
        return element

    def last(self):
        element = self.pending.pop(-1)
        print("\tDelivering %s from queue" % (element.identifier()))
        return element

    def __iter__(self):
        for element in self.pending:
            yield element

    def __len__(self):
        return len(self.pending)

    def remove(self, element):
        self.pending.remove(element)


# It should be possible to replace the BinarySearchJob in the script for this dummy class and have a dummy exp running
# Probably not working tho
class BinarySearchDummyJob(Job):
    def __init__(self, variables, remote_dir=None):
        super(BinarySearchDummyJob, self).__init__(variables, "")
        # This class extends the classical job putting some extra variables that define the relations
        self.parent = None
        self.child = []
        self.child_number = None
        self.level = None

        # Dummy vars for simulate a submission
        self.end_time = None

    # @property
    def identifier(self):
        return var_set_identifier(self)

    # @property
    def ancestors(self):
        to_return = []
        cv = self
        while cv.parent is not None:
            to_return.append(cv.parent)
            cv = cv.parent
        # return list(reversed(to_return))
        return list(to_return)

    def result(self):
        return self._result

    def submit(self):
        if self.submitted is False:
            self.submitted = True
            # Check if result already exists
            completed = self.is_completed
            if not completed:
                self.end_time = time.time() + 2

    @property
    def status(self):
        if self.end_time:
            if time.time() > self.end_time:
                return "COMPLETED", "00:00"
            else:
                return "RUNNING" , "00:00"
        else:
            return "PENDING", "00:00"

    def spawn_children(self):
        _set1, _set2 = divide_set(self.variable_set)
        for index, var_set in enumerate([_set1, _set2]):
            job = BinarySearchDummyJob(var_set, communicator=self.communicator, vault=self.vault)
            job.parent = self
            job.child_number = index
            self.child.append(job)
        return self.child


class BinarySearchJob(Job):
    def __init__(self, variables, communicator=None, vault=None):
        super(BinarySearchJob, self).__init__(variables, communicator, vault)
        # This class extends the classical job putting some extra variables that define the relations
        self.parent = None
        self.child = []
        self.child_number = None
        self.level = None
        self.banned_variables = []
        self.previous_hashes = []
        self.is_banned = False

    # @property
    def identifier(self):
        return var_set_identifier(self)

    def evaluate_success(self):


    # @property
    def ancestors(self):
        to_return = []
        cv = self
        while cv.parent is not None:
            to_return.append(cv.parent)
            cv = cv.parent
        # return list(reversed(to_return))
        return list(to_return)

    def descendants(self):
        to_return = []
        pre_list = [child for child in self.child]
        while len(pre_list) > 0:
            cv = pre_list.pop(0)
            for child in cv.child:
                if child not in pre_list:
                    pre_list.append(child)
            to_return.append(cv)
        return to_return

    def propagate_ban(self, _queue, _ban_list):
        self.is_banned = True
        for variable in self.variable_set:
            _ban_list.append(variable)

        # Ban the variables and upstream it
        print("Propagating BAN up %s!" % self.identifier())
        for ancestor in self.ancestors():
            if ancestor.remove_banned_variables(_ban_list):
                ancestor.is_completed = False
                ancestor._result = None
                ancestor.is_successful = None

        if self.parent:
            self.parent.check_child(_queue, _ban_list)

        # Is is really necessary?
        # Stop running descendants
        # for descendant in self.descendants():
        #     if descendant in queue.running:
        #         if descendant.hash == self.hash:
        #             descendant.cancel_remote()
        #             queue.running.remove(descendant)
        #             queue.completed.append(descendant)
        #             descendant.is_completed = True
        #             descendant._result = self._result
        #             descendant.is_successful = self.is_successful
        #             descendant.is_banned = True

    def propagate_successful_up(self, _queue, _ban_list):
        print("Propagating SUCCESSFUL up %s!" % self.identifier())
        if self.parent:
            if self.parent.both_child_successful and self.parent.is_successful is False and self.parent.banned_variables == []:
                self.parent.propagate_ban(_queue, _ban_list)
            else:
                _queue.add(self.parent)

    def spawn_children(self):
        if len(self.variable_set) >= 2:
            if self.child == []:
                _set1, _set2 = divide_set(self.variable_set)
                for index, var_set in enumerate([_set1, _set2]):
                    job = BinarySearchJob(var_set, communicator=self.communicator, vault=self.vault)
                    job.parent = self
                    job.child_number = index
                    job.level = self.level + 1
                    self.child.append(job)
                return True

        return False

    @property
    def both_child_successful(self):
        if self.child:
            all_successful = True
            for _child in self.child:
                if _child.is_successful is not True:
                    all_successful = False
            return all_successful
        else:
            return None

    @property
    def both_child_banned(self):
        if self.child:
            all_banned = True
            for _child in self.child:
                if _child.is_banned is not True:
                    all_banned = False
            return all_banned
        else:
            return None

    @property
    def child_one_successful_one_banned(self):
        if self.child:
            if self.child[0].is_successful and self.child[1].is_banned:
                return True
            elif self.child[1].is_successful and self.child[0].is_banned:
                return True
        return False

    @property
    def unsuccessful_descendant(self):
        for descendant in self.descendants():
            if descendant.is_successful is False and descendant.is_banned is False:
                return True
        return False

    def check_child(self, _queue, _ban_list):
        print("From %s checking child!!" % self.identifier())
        if self.child:
            if self.both_child_successful:
                if self.is_completed is True:
                    if self.is_successful is False:
                        self.propagate_ban(_queue=_queue, _ban_list=_ban_list)
                    else:
                        raise AssertionError("How can this happen?")
                else:
                    _queue.add(self)
            elif self.child_one_successful_one_banned:
                successful_child = [c for c in self.child if c.is_successful][0]
                if self.hash == successful_child.hash:
                    self.is_successful = True
                    self._result = successful_child._result
                    self.parent.check_child(_queue, _ban_list)
                else:
                    _queue.add(self)

            elif self.both_child_banned:
                self.propagate_ban(_queue=_queue, _ban_list=_ban_list)

    def remove_banned_variables(self, _ban_list):
        something_removed = False
        for variable in _ban_list:
            try:
                self.variable_set.remove(variable)
                self.banned_variables.append(variable)
                something_removed = True
            except ValueError:
                pass
        return something_removed


# Functions to create the initial sets
def divide_set(set0):
    assert len(set0) > 1
    set1 = []
    set2 = set0[:]
    for i in range(int(len(set0)/2)):
        set1.append(set2.pop(0))
    return set1, set2


# It was quite expensive to use this routine (2 min), now spawning the sets only if necessary
def generate_binary_tree_sets(input_set, job_class=BinarySearchDummyJob, communicator=None, vault=None):

    print_time()
    # While the sets still can be divided, keep creating new sets
    level = 0
    level_sets = {0: [input_set[:]]}
    control = True
    while control:

        level_sets[level+1] = []
        control = False
        for _set in level_sets[level]:
            # _set.level = level
            try:

                set_1, set_2 = divide_set(_set)
                level_sets[level+1].append(set_1)
                level_sets[level + 1].append(set_2)
                control = True
            except AssertionError:

                level_sets[level + 1].append(_set)

        level += 1
    del level_sets[level]

    print_time()
    # Put all sets (lists of identifiers) into a list of BinarySearchDummyJob objects
    _all_sets = []
    for level in level_sets.keys():
        for _set in level_sets[level]:
            _job = job_class(_set, communicator, vault)
            _job.level = level
            _all_sets.append(_job)
    print_time()
    sets_with_a_single_variable = [ var_set for var_set in _all_sets if len(var_set.variable_set) == 1]
    # Filter the list to remove repeated elements
    for variable in original_set:
        _counter = 0
        for var_set in sets_with_a_single_variable:
            if var_set.variable_set == [variable]:
                if _counter == 0:
                    _counter += 1
                else:
                    _all_sets.remove(var_set)
        if _counter != 1:
            raise AssertionError("This should not happen! Counter: %i" % _counter)
    del sets_with_a_single_variable

    print_time()
    # Establishing the parent set for each set
    for var_set1 in _all_sets:
        diff = 1000
        for var_set2 in _all_sets:
            if var_set1 is not var_set2:
                if set_1_contained_in_2(var_set1, var_set2):
                    if var_set1.parent is None:
                        var_set1.parent = var_set2
                        diff = set_difference(var_set1, var_set2)
                    else:
                        current_diff = set_difference(var_set1, var_set2)
                        if current_diff == 0:
                            continue
                        if current_diff < diff:
                            var_set1.parent = var_set2
                            diff = current_diff
    del var_set1, var_set2
    print_time()
    # Establishing the child of each set
    for var_set in _all_sets:
        if var_set.parent is not None:
            var_set.parent.child.append(var_set)
    print_time()
    for var_set in _all_sets:
        if len(var_set.child) > 2:
            raise AssertionError("This should not happen!")

    return _all_sets


def evaluate_set(_var_set, problematic_variables):
    # Dummy function to evaluate a set
    for var in problematic_variables:
        if var in _var_set.variable_set:
            _var_set._result = 1000
            return _var_set._result
    _var_set._result = 0
    return _var_set._result


def var_set_identifier(var_set):
    to_return = ""
    current_var_set = var_set
    while current_var_set.parent is not None:
        if len(to_return) > 100:
            raise AssertionError("Something strange happens here!")
        to_return += str(current_var_set.child_number)
        current_var_set = current_var_set.parent
    to_return = "l%02s: %s" % (str(var_set.level), to_return[::-1])
    return to_return


# Probably never used anymore
def solve_inconsistent_state(completed_jobs, _threshold, queue, ban_list, backup_queue):
    to_return = False
    # Look for all jobs in completed jobs
    for job in completed_jobs:
        # Check if the job and it's child are also completed
        if job.child:
            all_true = True
            for child in job.child:
                if child.is_completed is False:
                    all_true = False

            # If they are check if the problematic case where the parent fails but not any of the child occurs
            if all_true:
                failed = job.result() >= _threshold
                if failed:
                    none_failed = True
                    for child in job.child:
                        child_failed = child.result() >= _threshold
                        if child_failed:
                            none_failed = False
                    if none_failed:
                        # If it happens, ban the parent case and propagate this change up
                        job.propagate_ban(_queue=queue, _ban_list=ban_list, _running_jobs=[], _completed_jobs=completed_jobs)
                        for child in job.child:
                            backup_queue.add(child)

                        to_return = True
    return to_return


# Check if there's a job already running with the same hash
def job_already_running_with_same_hash(current_job, queue):
    for job in queue.running:
        if job.hash == current_job.hash:
            return True
    return False


# Check if there's a job already running with the same hash
def job_id_from_running_with_same_hash(current_job, queue):
    for job in queue.running:
        if job.hash == current_job.hash:
            return job.job_id
    return None


if __name__ == "__main__":
    
    # Parameters
    # remote_scratch = "/gpfs/scratch/pr1egi00/pr1egi23/ROMS-runs/"
    remote_scratch = "/gpfs/scratch/pr1egi00/pr1egi23/BinaryApproach/"

    # Accuracy threshold
    threshold = 10**-1
    # Defining different thresholds depending on the level
    threshold_by_level = {}
    for level in range(20):
        threshold_by_level[level] = threshold * (3. / 4.) ** level

    max_running_jobs = 10

    # Establish SSH connection to remote machine
    hostname = "mn1.bsc.es"
    user = "pr1egi23"
    remote = SSH(user=user, host=hostname, remote_scratch=remote_scratch)

    # Load variable database
    vault_path = "vault_with_processed_a1bo.pkl"
    vault = load_vault(vault_path)

    # Keeping the variables that have an id (it's precision can be changed in the code)
    variables = sorted([v for v in vault.variables if v.id is not None], key=lambda x: -x.error)

    # Original set including all the variables
    original_set = [v.id for v in variables]

    # An option is to use the previously known information to reduce the number of variables that will be processed
    # original_set = [v.id for v in variables if v.error < 10**-3]

    # Creating all possible sets (NOT USED ANYMORE, will be deleted soon)
    # all_sets = generate_binary_tree_sets(original_set, job_class=BinarySearchJob, communicator=remote, vault=vault)
    # for whole_set in all_sets:
    #     if whole_set.parent is None:
    #         break
    recovering_from_checkpoint = False
    if not recovering_from_checkpoint:
        # Creating the first set that will include all the variables
        whole_set = BinarySearchJob(original_set, communicator=remote, vault=vault)
        # Defining it's level to 0
        whole_set.level = 0

        # Initializing an empty list that will contain the variables that are banned
        ban_list = []

        # Initializing a counter to count all the simulations that have been done
        counter = 0

        # Initializing an empty queue for the execution
        queue = Queue()
        # Initializing an empty queue for the risky candidates (subsets that need to be banned but at some point may be
        # integrable into the solution)
        risky_candidates = Queue()

        # Add the whole set to the queue
        queue.add(whole_set)
    else:
        queue, whole_set, ban_list = recover_checkpoint(remote)
        counter = len(queue.completed)

    #
    # for job in whole_set.descendants():
    #     if job.is_banned and job.both_child_successful:
    #         print("%.2e  %.2e" % (job.result(), threshold_by_level[job.level]))
    #         for child in job.child:
    #             print("\t%s %.2e  %.2e" % (child.identifier(), child.result(), threshold_by_level[child.level]))
    #             print(child.result() < threshold_by_level[child.level])
    #
    # exit(0)

    # Fix current issue:
    apply_fixes = False
    if apply_fixes:
        queue.add(whole_set)
        whole_set._result = None

    # Define a condition to keep iterating with the refinement
    def condition(job):
        result = job.result()
        if result is None:
            return True
        else:
            if result > threshold:
                return True
            else:
                return False

    # Initialize a counter for the loops performed
    loops = 0
    # Start simulations
    while condition(whole_set):
        print("RESTARTING LOOP!!   : %i" % loops)
        print("Running jobs: %i\tJobs in queue: %i" % (len(queue.running), len(queue)))
        loops += 1

        print("SUBMIT PART_________________")
        # Submit part (Can be integrated into the queue class?)
        while len(queue.running) < max_running_jobs and len(queue) > 0:
            current_set = queue.first()
            if current_set in queue:
                raise AssertionError
            print("Submitting job %s with hash %i !! Elements still in queue: %i" % (current_set.identifier(), current_set.hash, len(queue)))

            if current_set.unsuccessful_descendant:
                print("Job has unsuccessful descendants and will be moved to completed: %s" % (
                      current_set.identifier()))
                current_set.is_successful = False
                current_set.completed = True
                queue.completed.append(current_set)

            elif job_already_running_with_same_hash(current_set, queue):
                current_set.job_id = job_id_from_running_with_same_hash(current_set, queue)
                current_set.submitted = True
                print("JOB ALREADY RUNNING WITH SAME HASH! %s" % current_set.job_id)
                # Append job to running experiments list
                queue.running.append(current_set)
            else:
                # Submit job
                current_set.submit()
                # Append job to running experiments list
                queue.running.append(current_set)

        # Control part
        print("CONTROL PART_________________")
        for current_set in queue.running:
            # Check the status of the running jobs
            status, elapsed_time = current_set.status
            print("Job %s %s with hash %i\t%s\t%s" % (current_set.job_id, current_set.identifier(), current_set.hash, status, elapsed_time))

            # If the job is completed
            if status == "COMPLETED":
                counter += 1
                # Move the job from running to completed
                print("Removing %s from running queue!" % current_set.identifier())
                queue.running.remove(current_set)
                queue.completed.append(current_set)

                # If the error is above the defined threshold
                successful = current_set.result() < threshold_by_level[current_set.level]
                print("Job completed %20s !: %i    SUCCESS:%s" % (current_set.identifier(), current_set.hash, successful))
                # If the job is not successful
                if not successful:
                    current_set.is_successful = False
                    # Try to spawn
                    if current_set.spawn_children():
                        # If possible, add the child to que queue
                        for child in current_set.child:
                            # If the set is already banned, it is not necessary to run it
                            if not child.is_banned:
                                queue.add(child)
                    # If the set can not be split, ban the set
                    else:
                        current_set.propagate_ban(queue, ban_list)

                # If otherwise the simulation was successful
                else:
                    current_set.is_successful = True
                    if current_set.parent:
                        current_set.parent.check_child(queue, ban_list)

            elif status == "FAILED":
                # Problematic! It might fail for other reasons!
                current_set.is_successful = False
                current_set.is_successful = False
                # Try to spawn
                if current_set.spawn_children():
                    # If possible, add the child to que queue
                    for child in current_set.child:
                        # If the set is already banned, it is not necessary to run it
                        if not child.is_banned:
                            queue.add(child)
                # If the set can not be split, ban the set
                else:
                    current_set.propagate_ban(queue, ban_list)

        # Saving state for plotting purposes
        state_filename = "state_%04i.pkl" % loops
        saving_state(queue, state_filename)
        saving_state(queue)

        # Assertions
        print("ASSERTION PART_________________")
        try:
            check_consistency(whole_set, queue, ban_list)
        except EmptyQueuesError:
            if condition(whole_set) is True:
                raise AssertionError("Empty queue and no running jobs!")
        if loops % 10 == 0:
            checkpoint(queue)
        wait = False
        if wait:
            if len(queue.running) > 0:
                print("Waiting _______________")
                for i in range(1):
                    print("%i" % (10-i))
                    time.sleep(1)

    print("F*****g completed!? Final set with %i variables and error %.2e" % (len(whole_set.variable_set), whole_set.result()))