import numpy as np


def load_var(_filename, _var_name):
    import numpy as np
    from netCDF4 import Dataset as NetCDFFile
    try:
        nc = NetCDFFile(_filename)
    except OSError:
        print("File %s not found!" % _filename)
        exit(1)
    # Load longitudes, latitudes and bathymetry data
    # longitudes = nc.variables['nav_lon'][:]
    # latitudes = nc.variables['nav_lat'][:]
    _data = nc.variables[_var_name]
    _data = np.array(_data)
    # To solve problem with _FillValue
    _data[_data == 1e20] = 0
    return _data


class DataFile:
    def __init__(self, file_name):
        self.filename = file_name
        self.vars = {}

    def load_var(self, var_name, mask=None):
        if mask is None:
            self.vars[var_name] = load_var(self.filename, var_name)
        else:
            data = load_var(self.filename, var_name)
            try:
                t, z, x, y = data.shape
            except ValueError:
                t, x, y = data.shape
                z = None

            mask = np.array([mask for i in range(t)])
            if z is None:
                mask = np.sum(mask, axis=2)
            mask = np.logical_not(mask)
            self.vars[var_name] = np.ma.array(data, mask=mask)

    def __getitem__(self, key):
        return self.vars[key]


def my_metric(predictions, targets):
    try:
        differences = predictions - targets  # the DIFFERENCEs.
    except ValueError:
        raise AssertionError

    differences_squared = differences ** 2  # the SQUAREs of ^

    if len(predictions.shape) == 4:
        mean_of_differences_squared = differences_squared.sum(axis=(3, 2, 1))  # the MEAN of ^
    elif len(predictions.shape) == 3:
        mean_of_differences_squared = differences_squared.sum(axis=(1, 2))  # the MEAN of ^
    else:
        raise ValueError("Variable has wrong dimensions")

    rmse_val = np.sqrt(mean_of_differences_squared)  # ROOT of ^

    return rmse_val


filename_1 = "%FILENAME1%"
filename_2 = "%FILENAME2%"

# Load Mask
mask_filename = "/gpfs/scratch/pr1egi00/pr1egi23/MP-Repo/a1an/scripts/mesh_mask.nc"
mask = DataFile(mask_filename)
mask.load_var("tmask")
mask = mask["tmask"]

var_names = ["thetao", "so", "heatc", "saltc"]

for var_name in var_names:
    # Variable used
    # var_name = "thetao"
    # Load reference file
    ref = DataFile(filename_1)
    ref.load_var(var_name, mask)

    compressed_ref = ref[var_name].compressed()
    iqr = np.percentile(compressed_ref, 75) - np.percentile(compressed_ref, 25)

    # Files to compare
    f = DataFile(filename_2)
    f.load_var(var_name, mask)

    try:
        # Compute the RMSE for each time
        f_rmse = list(my_metric(f[var_name], ref[var_name]))
        # Compute the maximum RMSE
        max_rmse = np.max(f_rmse)
        # Find the corresponding time
        max_index = f_rmse.index(max_rmse)
        # Compute the Inter Quartile Range for the corresponding time
        compressed_ref = ref[var_name][max_index].compressed()
        iqr = np.percentile(compressed_ref, 75) - np.percentile(compressed_ref, 25)
        result = max_rmse / iqr
    except AssertionError:
        result = 1000

    print("%s %.2e" % (var_name, result))
