import paramiko
import os
import stat


def remote_isdir(attr):
    return stat.S_ISDIR(attr.st_mode)


class SSH:
    def __init__(self, user, host, remote_scratch=""):
        # Init
        self.user = user
        self.host = host
        self._host_config_id = None
        self._ssh = None
        self._ssh_config = None
        self._user_config_file = None
        self._host_config = None
        self.sftp = None
        self.remote_scratch = remote_scratch

        # SSH
        self.connect()
        self.init_transport()

    def init_transport(self):
        self.sftp = self._ssh.open_sftp()

    def get(self, remote_path, local_path):
        if not self.sftp:
            self.init_transport()
        self.sftp.get(remote_path, local_path)

    def put(self, local_path, remote_path):
        if not self.sftp:
            self.init_transport()
        self.sftp.put(local_path, remote_path)

    def connect(self):
            """
            Creates ssh connection to host

            :return: True if connection is created, False otherwise
            :rtype: bool
            """
            try:
                self._ssh = paramiko.SSHClient()
                self._ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self._ssh_config = paramiko.SSHConfig()
                self._user_config_file = os.path.expanduser("~/.ssh/config")
                if os.path.exists(self._user_config_file):
                    with open(self._user_config_file) as f:
                        # noinspection PyTypeChecker
                        self._ssh_config.parse(f)
                self._host_config = self._ssh_config.lookup(self.host)
                if 'identityfile' in self._host_config:
                    self._host_config_id = self._host_config['identityfile']
                if 'proxycommand' in self._host_config:
                    self._proxy = paramiko.ProxyCommand(self._host_config['proxycommand'])
                    self._ssh.connect(self._host_config['hostname'], 22, username=self.user,
                                      key_filename=self._host_config_id, sock=self._proxy)
                else:
                    self._ssh.connect(self._host_config['hostname'], 22, username=self.user,
                                      key_filename=self._host_config_id)
                return True
            except IOError as e:
                print('Can not create ssh connection to {0}: {1}', self.host, e.strerror)
                return False

    def execute(self, command):
        return self._ssh.exec_command(command)

    def list_dir(self, dir_path):
        return self.sftp.listdir(dir_path)

